##
# Googler - Google API Library for Python
#
# Copyright (C) 2014 Christian Jurk <commx@commx.ws>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from .. import __version__ as lib_version

def get_user_agent():
    """
    Build a user agent string for HTTP requests.

    :return: User agent string
    """
    return 'Googler/%s' % lib_version

def get_headers():
    """
    Build a dict with default headers.

    :return: dict
    """
    d = {
        'User-agent': get_user_agent()
    }

    return d


##
# Googler - Google API Library for Python
#
# Copyright (C) 2014 Christian Jurk <commx@commx.ws>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from ..utils.http import get_headers

import requests

# Base URL for the Geocoding API
API_URL = 'maps.googleapis.com/maps/api/geocode'

class Geocode(object):
    def __init__(self, result):
        self.address_components = result.get('address_components')
        self.formatted_address = result.get('formatted_address')
        self.latitude = result['geometry']['location']['lat']
        self.longitude = result['geometry']['location']['lng']

    def __repr__(self):
        return '<Geocode: %f %f>' % (self.latitude, self.longitude)

class GeocodeResult(object):
    def __init__(self, data):
        self.status = data['status']
        self.error_message = None
        self.results = []

        if self.status != 'OK':
            self.error_message = data['error_message']

        for result in data['results']:
            self.results.append(Geocode(result))

    def __iter__(self):
        return self.results.__iter__()

    def __len__(self):
        if self.status != 'OK':
            return 0
        return len(self.results)

def get_geocode(address, api_key, format='json', use_tls=True):
    """
    Perform a Geocoding lookup (Latitude/Longitude).

    :param address: Address to geocode
    :param api_key: API key
    :param format: Output format. Can be "json" or "xml"
    :param use_tls: Specifies whether the request is made with https
    :return:
    """
    if format not in ('json', 'xml'):
        raise AttributeError('format argument must be "json" or "xml"')

    url = 'https' if use_tls else 'http'
    url += '://%s/%s' % (API_URL, format)
    headers = get_headers()
    params = {'address': address}

    try:
        r = requests.get(url, params=params, headers=headers)
    except requests.RequestException as e:
        return None
    else:
        if format == 'json':
            return GeocodeResult(r.json())
        else:
            raise NotImplementedError('xml support is currently not available')


##
# Googler - Google API Library for Python
#
# Copyright (C) 2014 Christian Jurk <commx@commx.ws>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from .. import geocoding

import json
import unittest

class TestEncryption(unittest.TestCase):
    """
    Test case to test Geocode response parsing.
    """
    def setUp(self):
        self.response = json.loads('''{
   "results" : [
      {
         "address_components" : [
            {
               "long_name" : "1600",
               "short_name" : "1600",
               "types" : [ "street_number" ]
            },
            {
               "long_name" : "Amphitheatre Pkwy",
               "short_name" : "Amphitheatre Pkwy",
               "types" : [ "route" ]
            },
            {
               "long_name" : "Mountain View",
               "short_name" : "Mountain View",
               "types" : [ "locality", "political" ]
            },
            {
               "long_name" : "Santa Clara",
               "short_name" : "Santa Clara",
               "types" : [ "administrative_area_level_2", "political" ]
            },
            {
               "long_name" : "California",
               "short_name" : "CA",
               "types" : [ "administrative_area_level_1", "political" ]
            },
            {
               "long_name" : "United States",
               "short_name" : "US",
               "types" : [ "country", "political" ]
            },
            {
               "long_name" : "94043",
               "short_name" : "94043",
               "types" : [ "postal_code" ]
            }
         ],
         "formatted_address" : "1600 Amphitheatre Pkwy, Mountain View, CA 94043, USA",
         "geometry" : {
            "location" : {
               "lat" : 37.42291810,
               "lng" : -122.08542120
            },
            "location_type" : "ROOFTOP",
            "viewport" : {
               "northeast" : {
                  "lat" : 37.42426708029149,
                  "lng" : -122.0840722197085
               },
               "southwest" : {
                  "lat" : 37.42156911970850,
                  "lng" : -122.0867701802915
               }
            }
         },
         "types" : [ "street_address" ]
      }
   ],
   "status" : "OK"
}''')

    def test_response_parsing(self):
        result = geocoding.GeocodeResult(self.response)

        for geocode in result.results:
            print(geocode)

        self.assertTrue(result)

if __name__ == '__main__':
    unittest.main()


##
# Googler - Google API Library for Python
#
# Copyright (C) 2014 Christian Jurk <commx@commx.ws>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from .. import mailhide

import unittest

class TestEncryption(unittest.TestCase):
    """
    Test case to test whether Encryption and Decryption for the mailhide
    API works correctly.
    """
    def setUp(self):
        self.private_key = 'deadbeefdeadbeefdeadbeefdeadbeef'
        self.email_address_list = (
            'x@example.com',
            'johndoe@example.com',
            'foobar.quite.long+custom@example.com',
        )

    def test_mailhide_encryption(self):
        for email in self.email_address_list:
            enc = mailhide._encrypt_email_address(email, self.private_key)
            dec = mailhide._decrypt_email_address(enc, self.private_key)
            self.assertEqual(email, dec)

if __name__ == '__main__':
    unittest.main()


##
# Googler - Google API Library for Python
#
# Copyright (C) 2014 Christian Jurk <commx@commx.ws>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from setuptools import setup, find_packages

setup(
    name='googler',
    version='1.0.0',
    author='Christian Jurk',
    author_email='commx@commx.ws',
    description=('Google API Library for Python',),
    license='Apache',
    keywords='google api recaptcha geocode geocoding',
    url='https://github.com/commx/googler',
    packages=find_packages(),
    install_requires=[
        'pycrypto',
        'requests'
    ],
    long_description='README.md',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries'
    ],
)


##
# Googler - Google API Library for Python
#
# Copyright (C) 2014 Christian Jurk <commx@commx.ws>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from Crypto.Cipher import AES

import base64

def get_html(email, private_key, public_key, use_tls=True, label=None):
    """
    Get Mailhide HTML code for a specific E-mail address.

    :param email: E-mail address to use for mailhide
    :param private_key: Private key
    :param public_key: Public key
    :param use_tls: Specifies whether to use https for the URL
    :param label: Optional - if present, it will display that label instead of a truncated E-mail address
    :return: Mailhide HTML code
    """
    url = get_url(email, private_key, public_key, use_tls)

    html_anchor_begin = '''<a href="%(href)s" onclick="window.open('%(href)s', '', 'toolbar=0,scrollbars=0,''' \
            '''location=0,statusbar=0,menubar=0,resizable=0,width=500,height=300');return false;" ''' \
            '''title="Reveal this e-mail address">'''
    html_anchor_end = '</a>'

    context = {
        'a_begin': html_anchor_begin % {
            'href': url
        },
        'a_end': html_anchor_end
    }

    if not label:
        email_parts = email.split('@', 2)
        username = email_parts[0]
        host = email_parts[1]
        prefix = ''

        if len(username) in range(1, 4 + 1):
            prefix = username[:2]
        elif len(username) in range(5, 6 + 1):
            prefix = username[:3]
        elif len(username) >= 7:
            prefix = username[:4]

        context.update({
            'prefix': prefix,
            'host': host
        })

        return '%(prefix)s%(a_begin)s…%(a_end)s@%(host)s' % context
    else:
        context.update({
            'label': label
        })

        return '%(a_begin)s%(label)s%(a_end)s' % context

def get_url(email, private_key, public_key, use_tls=True):
    """
    Get Mailhide URL for a specific E-mail address.

    :param email: E-mail address to use for mailhide
    :param private_key: Private key
    :param public_key: Public key
    :param use_tls: Specifies whether to use https for the URL
    :return: Mailhide URL
    """
    enc = _encrypt_email_address(email, private_key)

    ctx = {
        'public_key': public_key,
        'encrypted_email': base64.urlsafe_b64encode(enc).decode('utf-8')
    }

    url = 'https' if use_tls else 'http'
    url += '://www.google.com/recaptcha/mailhide/d?k=%(public_key)s&c=%(encrypted_email)s' % ctx

    return url

def _decrypt_email_address(cipher_text, private_key):
    """
    Decrypt a encrypted E-mail address.
    :param cipher_text: Encrypted E-mail address
    :param private_key: Private key
    :return: E-mail address in plain text
    """
    iv = bytes([0] * 16)
    key = base64.b16decode(private_key, casefold=True)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    plain_text = _unpad(cipher.decrypt(cipher_text))

    return plain_text.decode('utf-8')

def _encrypt_email_address(email, private_key):
    """
    Encrypt the E-mail address with AES-128-CBC.

    :param email: E-mail address
    :return:
    """
    padded_email = _pad(email.encode('utf-8'))
    key = base64.b16decode(private_key, casefold=True)

    # Create an initialization vector (16 times 0x00)
    iv = bytes([0] * 16)
    cipher = AES.new(key, AES.MODE_CBC, iv)

    return cipher.encrypt(padded_email)

def _pad(s):
    """
    A helper function for padding a message for use with AES.

    :param s: Source string
    :return: Padded string
    """
    x = AES.block_size - len(s) % AES.block_size
    return s + (bytes([x]) * x)

def _unpad(s):
    """
    A helper function to remove padding from a previously padded message.
    :param s: Padded string
    :return: Unpadded string
    """
    return s[:-s[-1]]


##
# Googler - Google API Library for Python
#
# Copyright (C) 2014 Christian Jurk <commx@commx.ws>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from ..utils.http import get_user_agent

import requests

# Base URL for the reCAPTCHA API
API_URL = 'www.google.com/recaptcha/api'

# HTML code for displaying the reCAPTCHA
DISPLAY_HTML = '''
<script type="text/javascript" src="%(api_url)s/challenge?k=%(public_key)s%(error_param)s">
</script>

<noscript>
  <iframe src="%(api_url)s/noscript?k=%(public_key)s%(error_param)s" height="300" width="500" frameborder="0"></iframe>
  <br/>
  <textarea name="recaptcha_challenge_field" rows="3" cols="40" required="required"></textarea>
  <input type="hidden" name="recaptcha_response_field" value="manual_challenge"/>
</noscript>
'''

# Error codes and their more verbose descriptions
ERROR_CODES = {
    'invalid-site-private-key': 'Unable to verify the private key',
    'invalid-request-cookie': 'The challenge parameter of the verify script was incorrect',
    'incorrect-captcha-sol': 'The CAPTCHA solution was incorrect',
    'captcha-timeout': 'The solution was received after the CAPTCHA timed out',
    'recaptcha-not-reachable': 'reCAPTCHA server currently not available'
}

class RecaptchaError(ValueError):
    """
    A base exception for all reCAPTCHA related errors.
    """
    def __init__(self, error_code):
        self.error_code = error_code

    def __str__(self):
        return ERROR_CODES.get(self.error_code)

class IncorrectRecaptchaSolution(RecaptchaError):
    """
    Raised when an invalid reCAPTCHA solution was provided.
    """
    pass

class RecaptchaResponse(object):
    """
    A reCAPTCHA response object.

    You can perform boolean testing on these objects for check whether
    the response was valid or not.
    """
    def __init__(self, is_valid, error_code=None):
        self.is_valid = is_valid
        self.error_code = error_code

    def __bool__(self):
        return bool(self.is_valid)

    @property
    def error_message(self):
        return ERROR_CODES.get(self.error_code)

def get_html(public_key, use_tls=True, error=None):
    """
    Get HTML code to display a reCAPTCHA.

    :param public_key: Public API key
    :param use_tls: Specifies whether the request is made with https
    :param error: A optional error message to display
    :return: HTML code for reCAPTCHA
    """
    error_param = ''

    if error:
        error_param = '&error=%s' % error

    context = {
        'api_url': _build_api_url(use_tls),
        'public_key': public_key,
        'error_param': error_param
    }

    return DISPLAY_HTML % context

def verify(challenge, response, private_key, remote_ip, use_tls=True):
    """
    Verify the reCAPTCHA response.

    On success, this function returns True. Otherwise, InvalidRecaptchaSolution
    is raised.

    :param challenge: The value of recaptcha_challenge_field from the form
    :param response: The value of recaptcha_response_field from the form
    :param private_key: Private API key
    :param remote_ip: User's IP address
    :param use_tls: Specifies whether the request is made with https
    :return: True on success
    """
    url = '%s/verify' % _build_api_url(use_tls)
    payload = {
        'privatekey': private_key,
        'remoteip': remote_ip,
        'challenge': challenge,
        'response': response
    }

    headers = _build_headers()

    try:
        r = requests.post(url, data=payload, headers=headers)
    except requests.RequestException as e:
        raise RecaptchaError('recaptcha-not-reachable')
    else:
        response = r.text.split('\n')

        if response[0].strip() == 'true':
            return True
        else:
            error_code = response[1].strip()

            if error_code == 'incorrect-captcha-sol':
                raise IncorrectRecaptchaSolution(error_code)
            else:
                raise RecaptchaError(error_code)

def _build_api_url(use_tls=True):
    """
    Build the API URL.

    :param use_tls: Specifies whether https is used for the URL
    :return: Full API URL
    """
    scheme = 'https' if use_tls else 'http'
    return '%s://%s' % (scheme, API_URL)

def _build_headers():
    """
    Build a headers dictionary containing pre-defined headers for HTTP requests.

    :return: Headers dictionary
    """
    headers = {
        'Content-type': 'application/x-www-form-urlencoded',
        'User-agent': get_user_agent()
    }

    return headers